﻿using System;
using System.Collections.Generic;

#nullable enable

namespace ExpressionEvaluatorI_2024 {

	public abstract class Expression {
		public abstract int Evaluate();
	}

	// Value leaf node infrastracture:

	public abstract class ValueExpression : Expression {
		public abstract int Value { get; }

		public sealed override int Evaluate() {
			return Value;
		}
	}

	// Value node types:

	public sealed class ConstantExpression : ValueExpression {
		private readonly int _value;

		public ConstantExpression(int value) {
			_value = value;
		}

		public override int Value => _value;
	}

	// Operator infrastructure:

	public abstract class OperatorExpression : Expression {
		public abstract bool AddOperand(Expression operand);
	}

	public abstract class UnaryExpression : OperatorExpression {
		private Expression? _operand;

		public Expression Operand {
			get { return _operand!; }
			set { _operand = value; }
		}

		public sealed override bool AddOperand(Expression operand) {
			_operand ??= operand;
			return true;
		}

		public sealed override int Evaluate() {
			return EvaluateUnaryOperator(_operand!.Evaluate());
		}

		protected abstract int EvaluateUnaryOperator(int value);
	}

	public abstract class BinaryExpression : OperatorExpression {
		private Expression? _leftOperand, _rightOperand;

		public Expression LeftOperand {
			get { return _leftOperand!; }
			set { _leftOperand = value; }
		}

		public Expression RightOperand {
			get { return _rightOperand!; }
			set { _rightOperand = value; }
		}

		public sealed override bool AddOperand(Expression operand) {
			if (_leftOperand == null) {
				_leftOperand = operand;
				return false;
			} else {
				_rightOperand ??= operand;
				return true;
			}
		}

		public sealed override int Evaluate() {
			return EvaluateBinaryOperator(_leftOperand!.Evaluate(), _rightOperand!.Evaluate());
		}

		protected abstract int EvaluateBinaryOperator(int leftValue, int rightValue);
	}

	// Binary operators:

	public sealed class PlusExpression : BinaryExpression {
		protected override int EvaluateBinaryOperator(int leftValue, int rightValue) {
			return checked(leftValue + rightValue);
		}
	}

	public sealed class MinusExpression : BinaryExpression {
		protected override int EvaluateBinaryOperator(int leftValue, int rightValue) {
			return checked(leftValue - rightValue);
		}
	}

	public sealed class MultiplyExpression : BinaryExpression {
		protected override int EvaluateBinaryOperator(int leftValue, int rightValue) {
			return checked(leftValue * rightValue);
		}
	}

	public sealed class DivideExpression : BinaryExpression {
		protected override int EvaluateBinaryOperator(int leftValue, int rightValue) {
			return checked(leftValue / rightValue);    // Can generate DivideByZeroException
		}
	}

	// Unary operators:

	public sealed class UnaryMinusExpression : UnaryExpression {
		protected override int EvaluateUnaryOperator(int value) {
			return checked(-value);
		}
	}

	// Prefix notation parser:

	public static class ExpressionParser {
		private readonly static char[] s_WhitespaceSeparators = { ' ' };

		private static Expression? CreateExpressionFromToken(string token) => token switch {
			// Binary operators:
			"+" => new PlusExpression(),
			"-" => new MinusExpression(),
			"*" => new MultiplyExpression(),
			"/" => new DivideExpression(),

			// Unary operators:
			"~" => new UnaryMinusExpression(),

			// Values:
			_ when int.TryParse(token, out int value) => 
				value >= 0 ?
					new ConstantExpression(value)
					:
					null,	// Negative values are not valid input

			// Invalid token format:
			_ => null   
		};

		public static Expression? ParsePrefixExpression(string exprString) {
			string[] tokens = exprString.Split(s_WhitespaceSeparators, StringSplitOptions.RemoveEmptyEntries);

			Expression? result = null;
			var unresolved = new Stack<OperatorExpression>();
			foreach (string token in tokens) {
				if (result != null) {
					// We correctly parsed the whole tree, but there was at least one more unprocessed token left.
					// This implies incorrect input, thus return null.

					return null;
				}

				switch (CreateExpressionFromToken(token)) {
					case null:
						return null;    // Invalid token format = incorrect input, thus return null.

					case OperatorExpression operatorExpr:
						unresolved.Push(operatorExpr);
						break;

					case ValueExpression valueExpr:

						Expression expr = valueExpr;
						while (unresolved.Count > 0) {
							OperatorExpression operatorExpr = unresolved.Peek();
							if (operatorExpr.AddOperand(expr)) {
								unresolved.Pop();
								expr = operatorExpr;
							} else {
								break;
							}
						}

						if (unresolved.Count == 0) {
							result = expr;	// Candidate for result, just need to check no more tokens left
						}

						break;

					default:
						throw new InvalidOperationException($"Unexpected expression type instance returned from {nameof(CreateExpressionFromToken)}() method.");
				}
			}

			return result;
		}
	}

	// Assignment solution:

	class Program {
		static void Main(string[] args) {
			string inputLine = Console.ReadLine()!;	// Assignment: at least one line is always present on standard input, we do not need to handle empty input

			Expression? expr = ExpressionParser.ParsePrefixExpression(inputLine);

			if (expr == null) {
				Console.WriteLine("Format Error");
				return;
			}

			try {
				int value = expr.Evaluate();
				Console.WriteLine(value.ToString());
			} catch (DivideByZeroException) {
				Console.WriteLine("Divide Error");
			} catch (OverflowException) {
				Console.WriteLine("Overflow Error");
			}
		}
	}

}
